"""
pyart.retrieve.simple_moment_calculations
=========================================

Simple moment calculations.

.. autosummary::
    :toctree: generated/

    compute_ccor
    calculate_snr_from_reflectivity
    compute_radial_noise_hs
    compute_radial_noise_ivic
    compute_noisedBZ
    compute_vol_refl
    compute_signal_power
    compute_rcs_from_pr
    compute_rcs
    compute_snr
    compute_l
    compute_cdr
    compute_bird_density
    calculate_velocity_texture
    atmospheric_gas_att
    get_coeff_attg
    _coeff_attg_table

"""
from warnings import warn
from copy import deepcopy

import numpy as np
from scipy import ndimage

from ..config import get_metadata, get_field_name
from ..core.transforms import antenna_to_cartesian
from .echo_class import get_freq_band
from ..util import angular_texture_2d, estimate_noise_hs74
from ..util import estimate_noise_ivic13, ivic_pct_table
from ..util import ivic_flat_reg_var_max_table, ivic_flat_reg_wind_len_table
from ..util import ivic_snr_thr_table


def compute_ccor(radar, filt_field=None, unfilt_field=None, ccor_field=None):
    """
    Computes the clutter correction ratio (CCOR), i.e. the ratio between the
    signal without Doppler filtering and the signal with Doppler filtering

    Parameters
    ----------
    radar : Radar
        Radar object
    filt_field, unfilt_field : str
        Name of Doppler filtered and unfiltered fields
    ccor_field : str
        Name of the CCOR field

    Returns
    -------
    ccor_dict : field dictionary
        Field dictionary containing the CCOR

    """
    # parse the field parameters
    if filt_field is None:
        filt_field = get_field_name('reflectivity')
    if unfilt_field is None:
        unfilt_field = get_field_name('unfiltered_reflectivity')
    if ccor_field is None:
        ccor_field = get_field_name('clutter_correction_ratio_hh')

    ccor_dict = get_metadata(ccor_field)
    ccor_dict['data'] = (
        radar.fields[unfilt_field]['data']-radar.fields[filt_field]['data'])

    return ccor_dict


def calculate_snr_from_reflectivity(
        radar, refl_field=None, snr_field=None, toa=25000.):
    """
    Calculate the signal to noise ratio, in dB, from the reflectivity field.

    Parameters
    ----------
    radar : Radar
        Radar object from which to retrieve reflectivity field.
    refl_field : str, optional
        Name of field in radar which contains the reflectivity.
        None will use the default field name in the Py-ART configuration file.
    snr_field : str, optional
        Name to use for snr metadata. None will use the default field name
        in the Py-ART configuration file.
    toa : float, optional
        Height above which to take noise floor measurements, in meters.

    Returns
    -------
    snr : field dictionary
        Field dictionary containing the signal to noise ratio.

    """
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if snr_field is None:
        snr_field = get_field_name('signal_to_noise_ratio')

    range_grid = np.meshgrid(radar.range['data'],
                             np.ma.ones(radar.time['data'].shape))[0] + 1.0

    # remove range scale.. This is basically the radar constant scaled dBm
    pseudo_power = (radar.fields[refl_field]['data'] -
                    20.0*np.log10(range_grid / 1000.0))

    # Noise floor estimate
    # 25km.. should be no scatterers, not even planes, this high
    # we could get undone by AP though.. also sun
    rg, azg = np.meshgrid(radar.range['data'], radar.azimuth['data'])
    rg, eleg = np.meshgrid(radar.range['data'], radar.elevation['data'])
    _, _, z = antenna_to_cartesian(rg / 1000.0, azg, eleg)  # XXX: need to fix

    points_above = np.where(z > toa)
    noise_floor_estimate = pseudo_power[points_above].mean()

    snr_dict = get_metadata(snr_field)
    snr_dict['data'] = pseudo_power - noise_floor_estimate
    return snr_dict


def compute_radial_noise_hs(radar, ind_rmin=0, nbins_min=1, max_std_pwr=2.,
                            pwr_field=None, noise_field=None,
                            get_noise_pos=False):
    """
    Computes radial noise in dBm from signal power using the algorithm
    from Hildebrand and Sekhon 1974

    Parameters
    ----------
    radar: radar object
        radar object containing the signal power in dBm
    ind_rmin: int
        index of the gate nearest to the radar where start looking for noisy
        gates
    nbins_min: int
        min number of noisy gates to consider the estimation valid
    max_std_pwr: float
        max standard deviation of the noise power to consider the noise valid
    pwr_field: str
        Name of the input signal power field
    noise_field: str
        name of the noise field to use
    get_noise_pos : bool
        If true an additional field with gates containing noise according to
        the algorithm is produced

    Returns
    -------
    noise_dict : dict
        the noise field in dBm
    noise_pos_dict : dict or None
        a dictionary containing a field where the gates with noise are set
        to 2 and those without are set to 1 (0 reserved)

    References
    ----------
    P. H. Hildebrand and R. S. Sekhon, Objective Determination of the Noise
    Level in Doppler Spectra. Journal of Applied Meteorology, 1974, 13,
    808-811.

    """
    # parse the field parameters
    if pwr_field is None:
        pwr_field = get_field_name('signal_power_hh')
    if noise_field is None:
        noise_field = get_field_name('noisedBm_hh')

    # extract fields from radar
    radar.check_field_exists(pwr_field)
    pwr = radar.fields[pwr_field]['data']
    pwr_mw = np.ma.power(10., 0.1*pwr)
    noise = np.ma.masked_all((radar.nrays, radar.ngates))
    if get_noise_pos:
        noise_pos = np.ma.zeros((radar.nrays, radar.ngates), dtype=np.uint8)
        noise_pos[np.ma.getmaskarray(pwr_mw)] = np.ma.masked
    for ray in range(radar.nrays):
        pwr_valid = pwr_mw[ray, ind_rmin:].compressed()
        mean, _, _, nnoise = estimate_noise_hs74(pwr_valid)
        if nnoise < nbins_min:
            continue
        noise[ray, :] = mean
        if get_noise_pos:
            ind_noise = np.argsort(pwr_valid)[0:nnoise]
            is_valid = np.logical_not(
                np.ma.getmaskarray(noise_pos[ray, ind_rmin:]))
            ind_valid = is_valid.nonzero()[0]
            noise_pos[ray, ind_rmin+ind_valid[ind_noise]] = 1

    noise = 10.*np.ma.log10(noise)
    noise_dict = get_metadata(noise_field)
    noise_dict['data'] = noise

    noise_pos_dict = None
    if get_noise_pos:
        if 'hh' in noise_field:
            noise_pos_field = 'noise_pos_h'
        else:
            noise_pos_field = 'noise_pos_v'
        noise_pos_dict = get_metadata(noise_pos_field)
        noise_pos_dict['data'] = noise_pos+1

    return noise_dict, noise_pos_dict


def compute_radial_noise_ivic(radar, npulses_ray=30, ngates_min=800,
                              iterations=10, pwr_field=None, noise_field=None,
                              get_noise_pos=False):
    """
    Computes radial noise in dBm from signal power using the algorithm
    described in Ivic et al. 2013

    Parameters
    ----------
    radar: radar object
        radar object containing the signal power in dBm
    npulses_ray : int
        Default number of pulses used in the computation of the ray. If the
        number of pulses is not in radar.instrument_parameters this will be
        used instead
    ngates_min: int
        minimum number of gates with noise to consider the retrieval valid
    iterations: int
        number of iterations in step 7
    pwr_field: str
        Name of the input signal power field
    noise_field: str
        name of the noise field to use
    get_noise_pos : bool
        If true an additional field with gates containing noise according to
        the algorithm is produced

    Returns
    -------
    noise_dict : dict
        the noise field in dBm
    noise_pos_dict : dict
        the position of the noisy gates
    get_noise_pos : bool
        If true an additional field with gates containing noise according to
        the algorithm is produced

    References
    ----------
    I.R. Ivic, C. Curtis and S.M. Torres, Radial-Based Noise Power Estimation
    for Weather Radars. Journal of Atmospheric and Oceanic Technology, 2013,
    30, 2737-2753.

    """
    # parse the field parameters
    if pwr_field is None:
        pwr_field = get_field_name('signal_power_hh')
    if noise_field is None:
        noise_field = get_field_name('noisedBm_hh')

    # extract fields from radar
    radar.check_field_exists(pwr_field)
    pwr_w = 1e-3*np.ma.power(10., 0.1*radar.fields[pwr_field]['data'])

    noise = np.ma.masked_all((radar.nrays, radar.ngates))
    if get_noise_pos:
        noise_pos = np.ma.zeros((radar.nrays, radar.ngates), dtype=np.uint8)
        noise_pos[np.ma.getmaskarray(pwr_w)] = np.ma.masked

    # get number of pulses per ray
    if radar.instrument_parameters is not None:
        if 'number_of_pulses' in radar.instrument_parameters:
            npulses = radar.instrument_parameters['number_of_pulses']['data']
        else:
            warn('Unknown number of pulses per ray. Default value ' +
                 str(npulses_ray)+' will be used for all rays')
            npulses = np.zeros(radar.nrays, dtype=int)+npulses_ray
    else:
        warn('Unknown number of pulses per ray. Default value ' +
             str(npulses_ray)+' will be used for all rays')
        npulses = np.zeros(radar.nrays, dtype=int)+npulses_ray

    # threshold for step 1:
    pct = ivic_pct_table(npulses)

    # threshold for step 2:
    flat_reg_wlen = ivic_flat_reg_wind_len_table(npulses)
    # we want an odd window
    for i in range(radar.nrays):
        if flat_reg_wlen[i] % 2 == 0:
            flat_reg_wlen[i] += 1
    flat_reg_var_max = ivic_flat_reg_var_max_table(npulses)

    # threshold for step 3:
    snr_thr = ivic_snr_thr_table(npulses)

    for ray, npuls in enumerate(npulses):
        mean, _, _, inds_noise = estimate_noise_ivic13(
            pwr_w[ray, :], pct=pct[ray], delay=2,
            flat_reg_wlen=flat_reg_wlen[ray],
            flat_reg_var_max=flat_reg_var_max[ray], snr_thr=snr_thr[ray],
            npulses=npuls, ngates_min=ngates_min, iterations=iterations,
            get_noise_pos=get_noise_pos)
        if mean is None:
            continue

        noise[ray, :] = mean
        if get_noise_pos:
            noise_pos[ray, inds_noise] = 1

    noise_dict = get_metadata(noise_field)
    noise_dict['data'] = 10.*np.ma.log10(noise)+30.

    noise_pos_dict = None
    if get_noise_pos:
        if 'hh' in noise_field:
            noise_pos_field = 'noise_pos_h'
        else:
            noise_pos_field = 'noise_pos_v'
        noise_pos_dict = get_metadata(noise_pos_field)
        noise_pos_dict['data'] = noise_pos+1

    return noise_dict, noise_pos_dict


def compute_noisedBZ(nrays, noisedBZ_val, _range, ref_dist,
                     noise_field=None):
    """
    Computes noise in dBZ from reference noise value.

    Parameters
    ----------
    nrays : int
        Number of rays in the reflectivity field.
    noisedBZ_val : float
        Estimated noise value in dBZ at reference distance.
    _range : np array of floats
        Range vector in m.
    ref_dist : float
        Reference distance in Km.
    noise_field : str, optional
        Name of the noise field.

    Returns
    -------
    noisedBZ : dict
        The noise field.

    """
    # parse the field parameters
    if noise_field is None:
        noise_field = get_field_name('noisedBZ_hh')

    noisedBZ_vec = noisedBZ_val+20.*np.ma.log10(1e-3*_range/ref_dist)

    noisedBZ = get_metadata(noise_field)
    noisedBZ['data'] = np.tile(noisedBZ_vec, (nrays, 1))

    return noisedBZ


def compute_vol_refl(radar, kw=0.93, freq=None, refl_field=None,
                     vol_refl_field=None):
    """
    Computes the volumetric reflectivity from the effective reflectivity
    factor

    Parameters
    ----------
    radar : Radar
        radar object
    kw : float
        water constant
    freq : None or float
        radar frequency
    refl_field : str
        name of the reflectivity used for the calculations
    vol_refl_field : str
        name of the volumetric reflectivity

    Returns
    -------
    vol_refl_dict : dict
        volumetric reflectivity and metadata in 10log10(cm^2 km^-3)

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if vol_refl_field is None:
        vol_refl_field = get_field_name('volumetric_reflectivity')

    # extract fields from radar
    radar.check_field_exists(refl_field)
    refl = radar.fields[refl_field]['data']

    # determine the parameters
    if freq is None:
        # get frequency from radar metadata
        if (radar.instrument_parameters is not None and
                'frequency' in radar.instrument_parameters):
            freq = radar.instrument_parameters['frequency']['data'][0]
        else:
            warn('Unable to compute volumetric reflectivity. ' +
                 'Unknown radar frequency')
            return None

    wavelen = 3e8/freq*1e2  # [cm]
    vol_refl = (
        1e3*np.power(np.pi, 5.)*kw*np.ma.power(10., 0.1*refl) /
        np.power(wavelen, 4.))

    vol_refl_dict = get_metadata(vol_refl_field)
    vol_refl_dict['data'] = 10.*np.log10(vol_refl)

    return vol_refl_dict


def compute_signal_power(radar, lmf=None, attg=None, radconst=None,
                         lrx=0., lradome=0., refl_field=None, pwr_field=None):
    """
    Computes received signal power OUTSIDE THE RADOME in dBm from a
    reflectivity field.

    Parameters
    ----------
    radar : Radar
        radar object
    lmf : float
        matched filter losses
    attg : float
        1-way gas attenuation
    radconst : float
        radar constant
    lrx : float
        receiver losses from the antenna feed to the reference point
        (positive value) [dB]
    lradome : float
        1-way losses due to the radome (positive value) [dB]
    refl_field : str
        name of the reflectivity used for the calculations
    pwr_field : str
        name of the signal power field

    Returns
    -------
    s_pwr_dict : dict
        power field and metadata

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if pwr_field is None:
        pwr_field = get_field_name('signal_power_hh')
        if refl_field.endswith('_vv'):
            pwr_field = get_field_name('signal_power_vv')

    # extract fields from radar
    radar.check_field_exists(refl_field)
    refl = radar.fields[refl_field]['data']

    # determine the parameters
    if lmf is None:
        if radar.radar_calibration is not None:
            if refl_field.endswith('_vv'):
                if 'matched_filter_loss_v' in radar.radar_calibration:
                    lmf = (
                        radar.radar_calibration['matched_filter_loss_v'][
                            'data'][0])
            elif 'matched_filter_loss_h' in radar.radar_calibration:
                lmf = (
                    radar.radar_calibration['matched_filter_loss_h'][
                        'data'][0])

        if lmf is None:
            warn('Unknown matched filter losses. Assumed 1 dB')
            lmf = 1.
    if attg is None:
        if (radar.radar_calibration is not None and
                'path_attenuation' in radar.radar_calibration):
            attg = radar.radar_calibration['path_attenuation']['data'][0]
        elif (radar.instrument_parameters is not None and
              'frequency' in radar.instrument_parameters):
            # assign coefficients according to radar frequency
            attg = get_coeff_attg(
                radar.instrument_parameters['frequency']['data'][0])
        else:
            attg = 0.0
            warn('Unknown 1-way gas attenuation. It will be set to 0')
    if radconst is None:
        # determine it from meta-data
        if radar.radar_calibration is not None:
            if refl_field.endswith('_vv'):
                if 'calibration_constant_vv' in radar.radar_calibration:
                    radconst = (
                        radar.radar_calibration[
                            'calibration_constant_vv']['data'][0])
            elif 'calibration_constant_hh' in radar.radar_calibration:
                radconst = (
                    radar.radar_calibration['calibration_constant_hh']['data'][0])

        if radconst is None:
            raise ValueError(
                'Radar constant unknown. ' +
                'Unable to determine the signal power')

    rng = radar.range['data']/1000.
    gas_att = 2.*attg*rng
    rangedB = 20.*np.ma.log10(rng)

    s_pwr = refl-rangedB-gas_att-radconst-lmf+lrx+lradome

    s_pwr_dict = get_metadata(pwr_field)
    s_pwr_dict['data'] = s_pwr

    return s_pwr_dict


def compute_rcs_from_pr(radar, lmf=None, attg=None, radconst=None,
                        tx_pwr=None, antenna_gain=None, lrx=0., ltx=0.,
                        lradome=0., freq=None, refl_field=None,
                        rcs_field=None, neglect_gas_att=False):
    """
    Computes the radar cross-section (assuming a point target) from radar
    reflectivity by first computing the received power and then the RCS from
    it.

    Parameters
    ----------
    radar : Radar
        radar object
    lmf : float
        matched filter losses. If None it will be obtained from the attribute
        radar_calibration of the radar object
    attg : float
        1-way gas attenuation
    radconst : float
        radar constant
    tx_pwr : float
        radar transmitted power [dBm]
    antenna_gain : float
        antenna gain [dB]. If None it will be obtain from the
        instrument_parameters attribute of the radar object
    lrx : float
        receiver losses from the antenna feed to the reference point
        (positive value) [dB]
    lradome : float
        1-way losses due to the radome (positive value) [dB]
    freq : float
        radar frequency [Hz]. If none it will be obtained from the radar
        metadata
    refl_field : str
        name of the reflectivity used for the calculations
    rcs_field : str
        name of the RCS field
    neglect_gas_att : bool
        Whether to neglect or not gas attenuation in the estimation of the
        RCS

    Returns
    -------
    rcs_dict : dict
        RCS field and metadata

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if rcs_field is None:
        rcs_field = get_field_name('radar_cross_section_hh')
        if refl_field.endswith('_vv'):
            rcs_field = get_field_name('radar_cross_section_vv')

    # determine the parameters
    rng = radar.range['data']  # [m]
    if antenna_gain is None:
        if radar.instrument_parameters is not None:
            if refl_field.endswith('_vv'):
                if 'radar_antenna_gain_v' in radar.instrument_parameters:
                    antenna_gain = (
                        radar.instrument_parameters['radar_antenna_gain_v'][
                            'data'][0])
            elif 'radar_antenna_gain_h' in radar.instrument_parameters:
                antenna_gain = (
                    radar.instrument_parameters['radar_antenna_gain_h'][
                        'data'][0])

        if antenna_gain is None:
            raise ValueError(
                'Antenna gain unknown. ' +
                'Unable to compute RCS')
    g_lin = np.power(10., 0.1*antenna_gain)

    if tx_pwr is None:
        # determine it from meta-data
        if radar.radar_calibration is not None:
            if refl_field.endswith('_vv'):
                if 'transmit_power_v' in radar.radar_calibration:
                    tx_pwr = (
                        radar.radar_calibration['transmit_power_v']['data'][0])
            elif 'transmit_power_h' in radar.radar_calibration:
                tx_pwr = (
                    radar.radar_calibration['transmit_power_h']['data'][0])

        if tx_pwr is None:
            raise ValueError(
                'Transmitted power unknown. ' +
                'Unable to compute RCS')
    if freq is None:
        # get frequency from radar metadata
        if (radar.instrument_parameters is not None and
                'frequency' in radar.instrument_parameters):
            freq = radar.instrument_parameters['frequency']['data'][0]
        else:
            raise ValueError(
                'Radar frequency unknown. ' +
                'Unable to compute RCS')

    # get received power OUTSIDE THE RADOME for reflectivity field [W]
    s_pwr = np.ma.power(10., 0.1*(compute_signal_power(
        radar, lmf=lmf, attg=attg, radconst=radconst, lrx=lrx,
        lradome=lradome, refl_field=refl_field, pwr_field=None)['data']-30.))

    wavelen = 3e8/freq  # [m]

    if neglect_gas_att:
        gas_att = 1.
    else:
        ELEV, RNG = np.meshgrid(
            radar.elevation['data'], rng/1000., indexing='ij')
        gas_att = np.power(10., 0.1*atmospheric_gas_att(freq, ELEV, RNG))

    tx_pwr_out = np.power(10., 0.1*(tx_pwr-ltx-lradome-30.))  # [W]

    rcs = 10*np.ma.log10(
        s_pwr*np.power(4*np.pi, 3.)*np.power(rng, 4)*np.power(gas_att, 2.) /
        (tx_pwr_out*np.power(g_lin, 2.)*np.power(wavelen, 2.)))

    rcs_dict = get_metadata(rcs_field)
    rcs_dict['data'] = rcs

    return rcs_dict


def compute_rcs(radar, kw2=0.93, pulse_width=None, beamwidth=None, freq=None,
                refl_field=None, rcs_field=None):
    """
    Computes the radar cross-section (assuming a point target) from radar
    reflectivity.

    Parameters
    ----------
    radar : Radar
        radar object
    kw2 : float
        water constant
    pulse_width : float
        pulse width [s]
    beamwidth : float
        beamwidth [degree]
    freq : float
        radar frequency [Hz]. If none it will be obtained from the radar
        metadata
    refl_field : str
        name of the reflectivity used for the calculations
    rcs_field : str
        name of the RCS field

    Returns
    -------
    rcs_dict : dict
        RCS field and metadata

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if rcs_field is None:
        rcs_field = get_field_name('radar_cross_section_hh')
        if refl_field.endswith('_vv'):
            rcs_field = get_field_name('radar_cross_section_vv')

    # extract fields from radar
    radar.check_field_exists(refl_field)
    refl_lin = np.ma.power(10., 0.1*radar.fields[refl_field]['data'])

    # determine the parameters
    rng = deepcopy(radar.range['data'])
    rng_mat = np.broadcast_to(
        rng.reshape(1, radar.ngates), (radar.nrays, radar.ngates))

    if freq is None:
        # get frequency from radar metadata
        if (radar.instrument_parameters is not None and
                'frequency' in radar.instrument_parameters):
            freq = radar.instrument_parameters['frequency']['data'][0]
        else:
            raise ValueError(
                'Radar frequency unknown. ' +
                'Unable to compute RCS')
    wavelen = 3e8/freq  # [m]
    if pulse_width is None:
        # get pulse width from radar metadata
        if (radar.instrument_parameters is not None and
                'pulse_width' in radar.instrument_parameters):
            pulse_width = (
                radar.instrument_parameters['pulse_width']['data'][0])
        else:
            raise ValueError(
                'Pulse width unknown. ' +
                'Unable to compute RCS')
    if beamwidth is None:
        # determine it from meta-data
        if refl_field.endswith('_vv'):
            if (radar.instrument_parameters is not None and
                    'radar_beam_width_v' in radar.instrument_parameters):
                beamwidth = (
                    radar.instrument_parameters['radar_beam_width_v']['data'][
                        0])
        elif (radar.instrument_parameters is not None and
              'radar_beam_width_h' in radar.instrument_parameters):
            beamwidth = (
                radar.instrument_parameters['radar_beam_width_h']['data'][0])

        if beamwidth is None:
            raise ValueError(
                'Antenna beamwidth unknown. ' +
                'Unable to compute RCS')
    beamwidth_rad = beamwidth*np.pi/180.

    rcs = 10*np.ma.log10(
        np.power(np.pi, 6.)*kw2*3e8*pulse_width*np.power(beamwidth_rad, 2.) *
        np.power(rng_mat, 2.)*refl_lin*1e-18 /
        (16.*np.log(2.)*np.power(wavelen, 4.)))

    rcs_dict = get_metadata(rcs_field)
    rcs_dict['data'] = rcs

    return rcs_dict


def compute_snr(radar, refl_field=None, noise_field=None, snr_field=None):
    """
    Computes SNR from a reflectivity field and the noise in dBZ.

    Parameters
    ----------
    radar : Radar
        Radar object
    refl_field : str, optional
        Name of the reflectivity field to use.
    noise_field : str, optional
        Name of the noise field to use.
    snr_field : str, optional
        Name of the SNR field.

    Returns
    -------
    snr : dict
        The SNR field.

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if noise_field is None:
        noise_field = get_field_name('noisedBZ_hh')
    if snr_field is None:
        snr_field = get_field_name('signal_to_noise_ratio')

    # extract fields from radar
    radar.check_field_exists(refl_field)
    radar.check_field_exists(noise_field)

    refl = radar.fields[refl_field]['data']
    noisedBZ = radar.fields[noise_field]['data']

    snr_data = refl-noisedBZ

    snr = get_metadata(snr_field)
    snr['data'] = snr_data

    return snr


def compute_l(radar, rhohv_field=None, l_field=None):
    """
    Computes Rhohv in logarithmic scale according to L=-log10(1-RhoHV).

    Parameters
    ----------
    radar : Radar
        Radar object.
    rhohv_field : str, optional
        Name of the RhoHV field to use.
    l_field : str, optional
        Name of the L field.

    Returns
    -------
    l : dict
        L field.

    """
    # parse the field parameters
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')
    if l_field is None:
        l_field = get_field_name('logarithmic_cross_correlation_ratio')

    # extract rhohv field from radar
    radar.check_field_exists(rhohv_field)
    rhohv = radar.fields[rhohv_field]['data']

    rhohv[rhohv >= 1.] = 0.9999
    l_data = -np.ma.log10(1.-rhohv)

    l = get_metadata(l_field)
    l['data'] = l_data

    return l


def compute_cdr(radar, rhohv_field=None, zdr_field=None, cdr_field=None):
    """
    Computes the Circular Depolarization Ratio.

    Parameters
    ----------
    radar : Radar
        Radar object.
    rhohv_field : str, optional
        Name of the RhoHV field.
    zdr_field : str, optional
        Name of the ZDR field.
    cdr_field : str, optional
        Name of the CDR field.

    Returns
    -------
    cdr : dict
        CDR field.

    """
    # parse the field parameters
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')
    if cdr_field is None:
        cdr_field = get_field_name('circular_depolarization_ratio')

    # extract fields from radar
    radar.check_field_exists(rhohv_field)
    radar.check_field_exists(zdr_field)

    rhohv = radar.fields[rhohv_field]['data']
    zdrdB = radar.fields[zdr_field]['data']

    zdr = np.ma.power(10., 0.1*zdrdB)

    cdr_data = (
        10.*np.ma.log10(
            (1.+1./zdr-2.*rhohv*np.ma.sqrt(1./zdr)) /
            (1.+1./zdr+2.*rhohv*np.ma.sqrt(1./zdr))))

    cdr = get_metadata(cdr_field)
    cdr['data'] = cdr_data

    return cdr


def compute_bird_density(radar, sigma_bird=11, vol_refl_field=None,
                         bird_density_field=None):
    """
    Computes the bird density from the volumetric reflectivity

    Parameters
    ----------
    radar : Radar
        radar object
    sigma_bird : float
        Estimated bird radar cross-section
    vol_refl_field : str
        name of the volumetric reflectivity used for the calculations
    bird_density_field : str
        name of the bird density field

    Returns
    -------
    bird_density_dict : dict
        bird density data and metadata [birds/km^3]

    """
    # parse the field parameters
    if vol_refl_field is None:
        vol_refl_field = get_field_name('volumetric_reflectivity')
    if bird_density_field is None:
        bird_density_field = get_field_name('bird_density')

    # extract fields from radar
    radar.check_field_exists(vol_refl_field)
    vol_refl = radar.fields[vol_refl_field]['data']

    bird_density = np.ma.power(10., 0.1*vol_refl)/sigma_bird

    bird_density_dict = get_metadata(bird_density_field)
    bird_density_dict['data'] = bird_density

    return bird_density_dict





def calculate_velocity_texture(radar, vel_field=None, wind_size=4, nyq=None,
                               check_nyq_uniform=True):
    """
    Derive the texture of the velocity field.

    Parameters
    ----------
    radar: Radar
        Radar object from which velocity texture field will be made.
    vel_field : str, optional
        Name of the velocity field. A value of None will force Py-ART to
        automatically determine the name of the velocity field.
    wind_size : int, optional
        The size of the window to calculate texture from. The window is
        defined to be a square of size wind_size by wind_size.
    nyq : float, optional
        The nyquist velocity of the radar. A value of None will force Py-ART
        to try and determine this automatically.
    check_nyquist_uniform : bool, optional
        True to check if the Nyquist velocities are uniform for all rays
        within a sweep, False will skip this check. This parameter is ignored
        when the nyq parameter is not None.

    Returns
    -------
    vel_dict: dict
        A dictionary containing the field entries for the radial velocity
        texture.

    """
    # Parse names of velocity field
    if vel_field is None:
        vel_field = get_field_name('velocity')

    # Allocate memory for texture field
    vel_texture = np.zeros(radar.fields[vel_field]['data'].shape)

    # If an array of nyquist velocities is derived, use different
    # nyquist velocites for each sweep in texture calculation according to
    # the nyquist velocity in each sweep.

    if nyq is None:
        # Find nyquist velocity if not specified
        nyq = [radar.get_nyquist_vel(i, check_nyq_uniform) for i in
               range(radar.nsweeps)]
        for i in range(0, radar.nsweeps):
            start_ray, end_ray = radar.get_start_end(i)
            inds = range(start_ray, end_ray)
            vel_sweep = radar.fields[vel_field]['data'][inds]
            vel_texture[inds] = angular_texture_2d(
                vel_sweep, wind_size, nyq[i])
    else:
        vel_texture = angular_texture_2d(
            radar.fields[vel_field]['data'], wind_size, nyq)
    vel_texture_field = get_metadata('velocity')
    vel_texture_field['long_name'] = 'Doppler velocity texture'
    vel_texture_field['standard_name'] = (
        'texture_of_radial_velocity' + '_of_scatters_away_from_instrument')
    vel_texture_field['data'] = ndimage.filters.median_filter(vel_texture,
                                                              size=(wind_size,
                                                                    wind_size))
    return vel_texture_field


def atmospheric_gas_att(freq, elev, rng):
    """
    Computes the one-way atmospheric gas attenuation [dB] according to the
    empirical formula in Doviak and Zrnic (1993) pp 44.
    This formula is valid for elev < 10 deg and rng < 200 km so values above
    these will be saturated to 10 deg and 200 km respectively

    Parameters
    ----------
    freq : float
        radar frequency [Hz]
    elev : float or array of floats
        elevation angle [deg]
    rng : float or array of floats. If array must have the same size as elev
        range [km]

    Returns
    -------
    latm : float or array of floats
        1-way gas attenuation [dB]

    """
    elev_aux = deepcopy(elev)
    rng_aux = deepcopy(rng)
    if np.isscalar(elev_aux):
        if elev_aux > 10.:
            elev_aux = 10.
    else:
        elev_aux[elev_aux > 10.] = 10.

    if np.isscalar(rng_aux):
        if rng_aux > 200.:
            rng_aux = 200.
    else:
        rng_aux[rng_aux > 200.] = 200.

    if not np.isscalar(elev_aux) and not np.isscalar(rng_aux):
        elev_size = np.size(elev_aux)
        rng_size = np.size(rng_aux)
        if elev_size != rng_size:
            raise ValueError(
                'Unable to compute gas attenuation field. ' +
                'radar elevation field size is '+str(elev_size) +
                ' and radar range field size is '+str(rng_size))

    # S-band atmospheric attenuation
    latm = (
        0.5*(0.4+3.45*np.exp(-elev_aux/1.8)) *
        (1-np.exp(-rng_aux/(27.8+154.*np.exp(-elev_aux/2.2)))))
    if freq > 12e9:
        # X-band
        latm *= 1.5
    elif 2e9 <= freq <= 12e9:
        # C-band
        latm *= 1.2

    return latm


def get_coeff_attg(freq):
    """
    get the 1-way gas attenuation for a particular frequency

    Parameters
    ----------
    freq : float
        radar frequency [Hz]

    Returns
    -------
    attg : float
        1-way gas attenuation

    """
    coeff_attg_dict = _coeff_attg_table()

    freq_band = get_freq_band(freq)
    if (freq_band is not None) and (freq_band in coeff_attg_dict):
        return coeff_attg_dict[freq_band]

    if freq < 2e9:
        freq_band_aux = 'S'
    elif freq > 12e9:
        freq_band_aux = 'X'

    warn('Radar frequency out of range. ' +
         'Coefficients only applied to S, C or X band. ' +
         freq_band + ' band coefficients will be used')

    return coeff_attg_dict[freq_band_aux]


def _coeff_attg_table():
    """
    defines the 1-way gas attenuation for each frequency band.

    Returns
    -------
    coeff_attg_dict : dict
        A dictionary with the coefficients at each band

    """
    coeff_attg_dict = dict()

    # S band
    coeff_attg_dict.update({'S': 0.0080})

    # C band
    coeff_attg_dict.update({'C': 0.0095})

    # X band
    coeff_attg_dict.update({'X': 0.0120})

    return coeff_attg_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
metranet python library
================

functions to read METRANET files in pure python, no other library required!

.. autosummary::
    :toctree: generated/

    read_polar
    read_product
    read_file
    _get_radar_site_info
    _nyquist_vel
    _selex2deg

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    RadarData
    PolarParser

"""


import sys
import struct
import os
import copy
from warnings import warn
import traceback

import numpy as np

from . import metranet_reader
from .pmfile_structure import MSWEEP_HEADER, MRAY_HEADER, MMOMENT_HEADER, MMOMENT_INFO_STRUCTURE
from .pmfile_structure import BYTE_SIZES
from .pmfile_structure import PRAY_HEADER, PMOMENTS
from .dn_to_float import float_mapping_m, float_mapping_p, nyquist_vel
from .lzw15 import decompress, readbytes_fh, unpackbyte

# fix for python3
if sys.version_info[0] == 3:
    def xrange(i):
        return range(i)

###############################################################################
#        Constants
###############################################################################


# For some reasons, the radar name is not encoded in a consistent way in the
# M binary files, this maps all names in files to a single character
RENAME_RADARS_M = {'Weissfluhgipfel':'W',
                   'Albis':'A ',
                   'L':'L',
                   'Dole':'D',
                   'P':'P'}
# In P files, radar name is three characters, map to single character
RENAME_RADARS_P = {'WEI':'W',
                   'ALB':'A',
                   'LEM':'L',
                   'DOL':'D',
                   'PPM':'P'}

# Dict to map moment names hardcoded in M files to terminology used
# in metranet C library
MOM_NAME_MAPPING = {'Z_V_CLUT':'ZVC',
                    'Z_CLUT':'ZHC',
                    'W':'WID',
                    'V':'VEL',
                    'PHIDP':'PHI',
                    'STAT1':'ST1',
                    'STAT2':'ST2',
                    'CLUT':'CLT',
                    'UZ':'ZH',
                    'UZ_V':'ZV',
                    'APH':'MPH'}

# Other way round, f.ex CLT --> CLUT
MOM_NAME_MAPPING_INV = dict(zip(MOM_NAME_MAPPING.values(),
                                MOM_NAME_MAPPING.keys()))

class RadarData():
    """
    A class for storing radar data.

    Attributes
    ----------
    header : dictionary
        file header (metadara valid for the whole sweep)
    pol_header : list
        list of ray metadata dictionaries
    data : dictionary
        dic of arrays containing data
    """

    def __init__(self, header=None, pol_header=None, data=None, scale=None):
        self.header = header
        self.pol_header = pol_header
        self.data = data
        self.scale = scale


class PolarParser():
    """
    A class for parsing metranet polar data

    Parameters
    ----------
    filename : str
        complete file path of the file to read

    Attributes
    ----------
    file_format : char
        P or M depending on the file format that was used
    file_type : char
        file type (H, M or L)
    bname : str
        basename (without path) of the file to read
    bytarray: memoryview of a bytearray
        binary data contained in the file

    """

    def __init__(self, filename):
        # check if it is the right file. Open it and read it
        bfile = os.path.basename(filename)

        supported_file = (bfile.startswith('MH') or bfile.startswith('PH') or
                          bfile.startswith('MS') or bfile.startswith('PM') or
                          bfile.startswith('ML') or bfile.startswith('PL'))

        if not supported_file:
            raise ValueError(
                """Only polar data files starting by MS, MH or ML or PM, PH, PL
                   are supported""")

        self.file_format = bfile[0] # M or P
        self.file_type = bfile[1] # H, M or L
        self.bname = bfile

        # Open binary file
        with open(filename, 'rb') as f:
            self.bytearray = memoryview(bytearray(f.read())) # bytearray

        self.read_pos = 0 # Current "cursor" position in file

        # Get endianness
        if self.file_format == 'P':
            self.endian_prefix = '>'
        elif self.file_format == 'M':
            if sys.byteorder == 'little':
                self.endian_prefix = '<'
            else:
                self.endian_prefix = '>'

    def parse(self, moments=None):
        """
            Parses a metranet binary file by calling the appropriate parser
            depending on the file type (M or P)

            Parameters
            ----------
            moments : list (optional)
                list of moments to get from the file, possible moments
                are ZH, ZV, ZDR, ST1, ST2, VEL, WID, PHI, RHO, CLT, ZVC, ZHC,
                MPH, default = ALL moments available in the file are read


            Returns
            -------
            out : RadarClass instance
                RadarClass instance containing the file data and metadata

        """
        self.read_pos = 0 # reset read position if needed

        out = None
        if self.file_format == 'P':
            out = self._parse_p(moments)
        elif self.file_format == 'M':
            out = self._parse_m(moments)

        return out


    def _parse_p(self, moments):
        """
            Parses a P format metranet binary file

            Parameters
            ----------
            moments : list
                list of moments to get from the file, possible moments
                are ZH, ZV, ZDR, ST1, ST2, VEL, WID, PHI, RHO, CLT, ZVC, ZHC,
                MPH, if none ALL moments available in the file are read


            Returns
            -------
            head : dict
                file metadata
            pol_header: list
                list of metadata dictionaries for every ray
            moments_data: dictionary
                dictionary containing the data in DN for every moment

        """

        # For P files we get available moment from pmfile_structure.py
        moments_avail = np.array(PMOMENTS['names'][self.file_type])
        nummoments = len(moments_avail)

        if moments is None:
            moments = moments_avail

        # Initialize outputs
        pol_header = []
        moments_data = {}
        for m in moments:
            moments_data[m] = []


        while self.read_pos != len(self.bytearray):
            # Get ray metadata
            ray = self._get_chunk(PRAY_HEADER)

            # Convert DN angles to float angles
            ray['startangle_az'] = _selex2deg(ray['startangle_az'])
            ray['startangle_el'] = _selex2deg(ray['startangle_el'])
            ray['endangle_az'] = _selex2deg(ray['endangle_az'])
            ray['endangle_el'] = _selex2deg(ray['endangle_el'])
            pol_header.append(ray)

            for i in range(nummoments): # After ray metadata get all moments
                ngates = ray['numgates']

                if moments_avail[i] in moments: # Check if this moment is required

                    # Check type of this particular moment (byte or short)
                    if moments_avail[i] in PMOMENTS['types'].keys():
                        data_type = PMOMENTS['types'][moments_avail[i]]
                    else:
                        data_type = 'B'

                    ffmt = self.endian_prefix + data_type

                    len_mom = ngates * BYTE_SIZES[data_type]
                    mom = np.ndarray((ngates,), ffmt,
                                     self.bytearray[self.read_pos:
                                                    self.read_pos + len_mom])
                    self.read_pos += len_mom
                    moments_data[moments_avail[i]].append(mom)
                else:
                    # Even if moment is not required, we need to update cursor
                    self.read_pos += len_mom

        # Create global file header to be consistent with M files
        head = {}
        head['antmode'] = pol_header[0]['antmode']
        head['radarname'] = pol_header[0]['scanid']
        head['moments'] = moments
        # Rename radar with appropriate dict if needed, otherwise get name from file
        if head['radarname'] in RENAME_RADARS_P.keys():
            head['radarname'] = RENAME_RADARS_P[head['radarname']]
        else:
            head['radarname'] = self.bname[2]
        head['startrange'] = pol_header[0]['startrange']
        head['currentsweep'] = pol_header[0]['currentsweep']
        head['gatewidth'] = pol_header[0]['gatewidth']

        radmetadata = _get_radar_site_info()
        radname = head['radarname']
        head['frequency'] = radmetadata[radname]['Frequency']
        head['radarheight'] = radmetadata[radname]['RadarHeight']
        head['radarlat'] = radmetadata[radname]['RadarLat']
        head['radarlon'] = radmetadata[radname]['RadarLon']

        return head, pol_header, moments_data

    def _parse_m(self, moments):
        """
            Parses a M format metranet binary file

            Parameters
            ----------
            moments : list
                list of moments to get from the file, possible moments
                are ZH, ZV, ZDR, ST1, ST2, VEL, WID, PHI, RHO, CLT, ZVC, ZHC,
                MPH, if none ALL moments available in the file are read


            Returns
            -------
            head : dict
                file metadata
            pol_header: list
                list of metadata dictionaries for every ray
            moments_data: dictionary
                dictionary containing the data in DN for every moment

        """

        head = self._get_chunk(MSWEEP_HEADER)
        if head['radarname'] in RENAME_RADARS_M.keys():
            head['radarname'] = RENAME_RADARS_M[head['radarname']] # make consistent names
        else: # get radar name from filename
            head['radarname'] = self.bname[2]

        head['frequency'] *= 10**9 # Be consistent with C-library where it is in hZ
        nummoments = head['nummoments']

        pol_header = []

        # Get all available moments
        moments_avail = [a['name'] for a in head['moments']]

        if moments is None:
            moments = moments_avail

        moments_data = {}
        for m in moments:
            moments_data[m] = []

        while self.read_pos != len(self.bytearray):
            pol = self._get_chunk(MRAY_HEADER)
            # Convert DN angles to float angles
            pol['startangle_az'] = _selex2deg(pol['startangle_az'])
            pol['startangle_el'] = _selex2deg(pol['startangle_el'])
            pol['endangle_az'] = _selex2deg(pol['endangle_az'])
            pol['endangle_el'] = _selex2deg(pol['endangle_el'])

            # Convert datetime to UTC + residue
            pol['datatime_residue'] = int(100 * ((pol['datatime']* 0.01) % 1))
            pol['datatime'] = int(0.01 * pol['datatime'])
            pol_header.append(pol)

            for i in range(nummoments):
                mom_header = self._get_chunk(MMOMENT_HEADER)

                len_mom = mom_header['datasize']

                if moments_avail[i] in moments:
                    size_moment_bytes = head['moments'][i]['num_bytes']

                    ngates = int(len_mom/size_moment_bytes) # nb of gates
                    ffmt = self.endian_prefix
                    if size_moment_bytes == 1:
                        ffmt += 'B'
                    else:
                        ffmt += 'H'

                    mom = np.ndarray((ngates,), ffmt,
                                     self.bytearray[self.read_pos:
                                                    self.read_pos + len_mom])
                    moments_data[moments_avail[i]].append(mom)
                    self.read_pos += len_mom
                else:
                    self.read_pos += len_mom

        return head, pol_header, moments_data

    def _get_chunk(self, file_info):
        """
        Parses part of the bytearray given the info about what data to
        expect

        Parameters
        ----------
        file_info : dict
            dictionary with three keys: 'names', 'len', 'type'
            containing the names, length and types of the variables that
            should be read from the binary file. These dictionaries are
            defined in the pmfile_structure.py file


        Returns
        -------
        dic_values : dict
            dictionary containing all the variables that were read from
            the file, with keys corresponding to the 'names' key in the
            file_info dict

        """

        # Read the bytearray byte by byte
        dic_values = {} # output dictionary

        for i in range(len(file_info['names'])):
            len_val = file_info['len'][i]
            name_val = file_info['names'][i]

            if isinstance(len_val, list):
                # If length in pmfile_structure.py is a list, generate
                # real length with previously read key
                len_val = len_val[0] * dic_values[len_val[1]]

            type_var = file_info['type'][i]
            ffmt = self.endian_prefix + "{}".format(int(len_val))
            ffmt += type_var

            offset = len_val * BYTE_SIZES[type_var]


            if name_val == 'moments':

                # M files only
                # For the moments structure some additional processing is needed
                val = []
                for j in range(dic_values['nummoments']):
                    # recursion!
                    val.append(self._get_chunk(MMOMENT_INFO_STRUCTURE))


            else:
                val = struct.unpack_from(
                    ffmt, self.bytearray[self.read_pos:self.read_pos + offset])

                # Convert to array if list
                if len(val) == 1:
                    val = val[0]
                else:
                    val = np.array(val)

                if type_var == 's': # For strings
                    # Strip null hexadecimal codes
                    val = val.decode('utf-8').replace('\x00', '')
                self.read_pos += offset

            dic_values[name_val] = val


        return dic_values


def read_polar(filename, moments=None, physic_value=True, masked_array=True,
               reorder_angles=True):
    """
    Reads a METRANET polar data file

    Parameters
    ----------
    radar_file : str
        file name
    moments : list
        List of moments to read, by default all are used
    physic_value : boolean
        If true returns the physical value. Otherwise the digital value
    masked_array : boolean
        If true returns a numpy masked array with NaN values masked. Otherwise
        returns a regular masked array with NaN values
    reorder_anges: boolean
        If true all recorded rays are sorted by ascending order of their angles
        In addition the scan is truncated to a maximum of 360 rays


    Returns
    -------
    ret_data : RadarData object
        An object containing the information read from the file.

    """


    parser = PolarParser(filename)
    head, pol_header, moments_data = parser.parse(moments)

    # deep-copy needed since we alter the keys during loop
    moments = copy.deepcopy(list(moments_data.keys()))

    for m in moments:
        # Find index of moment in the moments structure of the header
        moments_data[m] = np.array(
            moments_data[m], dtype=moments_data[m][0].dtype)

        if masked_array:
            if parser.file_type == 'L' and m in ['UZ','UZ_V','Z_V_CLUT',
                                                 'Z_CLUT']:
                mask = np.logical_or(
                    moments_data[m] == 0, moments_data[m] == 1)
            else:
                mask = moments_data[m] == 0
                
        if parser.file_format == 'M':
            idx_mom_head = [h['name'] == m
                                    for h in head['moments']].index(True)
            if m in MOM_NAME_MAPPING.keys():
                moments_data[MOM_NAME_MAPPING[m]] = moments_data.pop(m)
                m = MOM_NAME_MAPPING[m]

        if physic_value:
            if parser.file_format == 'M':
                moments_data[m] = float_mapping_m(
                        m, head['moments'][idx_mom_head],
                        pol_header[0]['datatime'], head['radarname'],
                        nyquist_vel(head['currentsweep'] - 1))[
                            moments_data[m]].astype(np.float32)
            else:
                moments_data[m] = float_mapping_p(
                        m, pol_header[0]['datatime'], head['radarname'],
                        nyquist_vel(head['currentsweep'] - 1))[
                            moments_data[m]].astype(np.float32) 
        # Rename moment if needed


        if masked_array:
            moments_data[m] = np.ma.array(moments_data[m], mask=mask)
        else:
            moments_data[m][mask] = np.nan

    if reorder_angles:
        # Reorder dependent angle in ascending order
        if head['antmode'] in [0, 2]:
            angles = np.array([ray['startangle_az'] for ray in pol_header])
        else:
            angles = np.array([ray['startangle_el'] for ray in pol_header])

        nray = len(angles)

        maxrays = min([nray, 360])
        angles = angles[-maxrays:]
        order = np.argsort(angles)

        tmp = [pol_header[i + nray - maxrays] for i in order]
        pol_header = tmp

        for m in moments_data.keys():
            moments_data[m] = moments_data[m][-360:][order]

    ret_data = RadarData(
        header=head, pol_header=pol_header, data=moments_data)
    return ret_data


def read_product(radar_file, physic_value=False, masked_array=False,
                 verbose=False):
    """
    Reads a METRANET cartesian data file

    Parameters
    ----------
    radar_file : str
        file name
    physic_value : boolean
        If true returns the physical value. Otherwise the digital value
    masked_array : boolean
        If true returns a numpy masked array with NaN values masked. Otherwise
        returns a regular masked array with NaN values
    verbose : boolean
        If true prints out extra information

    Returns
    -------
    ret_data : RadarData object
        An object containing the information read from the file. None if
        the file has not been properly read

    """
    ret_data = RadarData()
    prd_header = {'row': 0, 'column': 0}

    # read ASCII data
    if verbose:
        print("physic_value: ", physic_value)
        print("File %s: read ASCII" % radar_file)

    try:
        with open(radar_file, 'rb') as data_file:
            for t_line in data_file:
                line = t_line.decode("utf-8").strip('\n')
                if line.find('end_header') == -1:
                    data = line.split('=')
                    prd_header[data[0]] = data[1]
                else:
                    break

            # read BINARY data
            prdt_size = int(prd_header['column']) * int(prd_header['row'])
            if prdt_size < 1:
                print("Error, no size found row=%3d column=%3d" %
                      (prd_header['row'], prd_header['column']))
                return None

            if verbose:
                print("File %s: read BINARY data: expected %s bytes, " %
                      (radar_file, prdt_size), end='')
                print(prd_header)

            if int(prd_header['table_size']) != 0:
                prd_data_level = np.fromfile(
                    data_file, dtype=np.dtype('>f4'),
                    count=int(int(prd_header['table_size']) / 4))
            else:
                prd_data_level = []

            # Read and decompress data using lzw15
            prd_data_compressed_iter = readbytes_fh(data_file)
            prd_data_iter = decompress(prd_data_compressed_iter)

            prd_data_list = list()
            for k in prd_data_iter:
                prd_data_list.append(unpackbyte(k))

            prd_data = np.asarray(prd_data_list, dtype=np.ubyte)
            prd_data = np.reshape(
                prd_data, (int(prd_header['row']), int(prd_header['column'])))

        data_file.close()

    except OSError as ee:
        warn(str(ee))
        print("Unable to read file '%s'" % radar_file)
        return None

    # convert 0 at end of array with NAN
    conv_zero2nan = True

    nlevels = np.size(prd_data_level)
    if nlevels == 0:
        prd_data_level = np.arange(256, dtype=np.uint32)
    else:
        while conv_zero2nan:
            if nlevels == 0:
                conv_zero2nan = False
            elif prd_data_level[nlevels-1] == 0.0:
                prd_data_level[nlevels-1] = np.nan
            else:
                conv_zero2nan = False

    if verbose:
        print("Found %d bytes" % np.size(prd_data))
        print("prd_data_level[10] = %f" % prd_data_level[10])
        print("min/max prd_data: %d/%d" % (prd_data.min(), prd_data.max()))
        print("first 100 bytes", prd_data[0:100, 0])
        print("data level ", prd_data_level[0:10])

    if physic_value:
        ret_data.data = prd_data_level[prd_data]
        if masked_array:
            ret_data.data = np.ma.array(
                ret_data.data, mask=np.isnan(ret_data.data))
            ret_data.data = np.ma.masked_where(prd_data == 0, ret_data.data)
    else:
        ret_data.data = prd_data
        if masked_array:
            ret_data.data = np.ma.array(
                ret_data.data, mask=prd_data == 0)
    ret_data.header = prd_header
    ret_data.scale = prd_data_level

    return ret_data


def read_file(file, moment="ZH", physic_value=False, masked_array=False,
              reorder_angles=True, verbose=False):
    """
    Reads a METRANET data file

    Parameters
    ----------
    file : str
        file name
    moment : str
        moment name
    physic_value : boolean
        If true returns the physical value. Otherwise the digital value
    masked_array : boolean
        If true returns a numpy masked array with NaN values masked. Otherwise
        returns a regular masked array with NaN values
    reorder_angles : boolean
        If true angles are reordered
    verbose : boolean
        If true prints out extra information

    Returns
    -------
    ret_data : RadarData object
        An object containing the information read from the file

    """
    bfile = os.path.basename(file)

    if (bfile.startswith('PM') or bfile.startswith('PH') or
            bfile.startswith('PL') or bfile.startswith('MS') or
            bfile.startswith('MH') or bfile.startswith('ML')):
        # polar data from SITE (PH/PM/PL)
        if verbose:
            print("calling read_polar")
        ret = read_polar(
            file, moments=moment, physic_value=physic_value,
            masked_array=masked_array, reorder_angles=reorder_angles)
    else:
        # cartesian / CCS4 products
        if verbose:
            print("calling read_product")
        ret = read_product(
            file, physic_value=physic_value, masked_array=masked_array,
            verbose=verbose)

    return ret


def _get_radar_site_info(verbose=False):
    """
        return dictionary with radar'info

        Returns
        -------
        radar_def : dict
            dictionary containing radar site information
    """
    radar_def_load = False
    try:
        import yaml
        yaml_module = True
    except ImportError:
        yaml_module = False

    if yaml_module:
        metranet_yaml_file = "metranet.yaml"
        path_yaml_file = [
            "/store/msrad/python/library/radar/io/", "/opt/ccs4/python",
            "/proj/lom/python/library/radar/io"]
        for p in path_yaml_file:
            full_file = p + "/" + metranet_yaml_file
            if os.path.isfile(full_file):
                print("look for Radar_Site_info from %s" % full_file)
                try:
                    with open(full_file) as f:
                        radar_def = yaml.load(f)
                        radar_def_load = True
                        if verbose:
                            print("Read Radar_Site_info from %s" % full_file)
                except OSError as ee:
                    warn(str(ee))
                    traceback.print_exc()

    if not radar_def_load:
        # HardCoded definition
        print("**** HARD-CODED Radar site Value, may be not updated!!")
        # some values valid for all sites
        c_speed = 299792458  # m/s

        radar_default = {}

        radar_default['dwhname'] = 'undef'
        radar_default['RadarCHY'] = float('nan')
        radar_default['RadarCHX'] = float('nan')
        radar_default['FileId'] = 1179665477
        radar_default['Version'] = 0
        radar_default['RadarName'] = "undef"
        radar_default['ScanName'] = "undef"
        radar_default['RadarLat'] = float('nan')
        radar_default['RadarLon'] = float('nan')
        radar_default['RadarHeight'] = float('nan')
        radar_default['Frequency'] = float('nan')
        radar_default['WaveLength'] = float('nan')
        radar_default['PulseWidth'] = 0.5
        radar_default['SweepsOrder'] = (
            9, 7, 5, 3, 1, 19, 17, 15, 13, 11, 10, 8, 6, 4, 2, 20, 18, 16, 14,
            12)
        radar_default['NumMomentsPM'] = metranet_reader.NPM_MOM
        radar_default['NumMomentsPH'] = metranet_reader.NPH_MOM
        radar_default['NumMomentsPL'] = metranet_reader.NPL_MOM

        radar_def = {}

        rname = 'A'
        radar_def[rname] = radar_default.copy()
        radar_def[rname]['dwhname'] = "ALB"
        radar_def[rname]['RadarName'] = "Albis"
        radar_def[rname]['RadarCHY'] = 681201
        radar_def[rname]['RadarCHX'] = 237604
        radar_def[rname]['RadarLat'] = 47.284333
        radar_def[rname]['RadarLon'] = 8.512000
        radar_def[rname]['RadarHeight'] = 938.0
        radar_def[rname]['ScanName'] = "1095516672"
        radar_def[rname]['Frequency'] = 5450e6
        radar_def[rname]['WaveLength'] = (
            c_speed/radar_def[rname]['Frequency']*1e2)

        rname = 'D'
        radar_def[rname] = radar_default.copy()
        radar_def[rname]['dwhname'] = "DOL"
        radar_def[rname]['RadarName'] = "Dole"
        radar_def[rname]['RadarCHY'] = 497057
        radar_def[rname]['RadarCHX'] = 142408
        radar_def[rname]['RadarLat'] = 46.425113
        radar_def[rname]['RadarLon'] = 6.099415
        radar_def[rname]['RadarHeight'] = 1682.0
        radar_def[rname]['ScanName'] = "1146047488"
        radar_def[rname]['Frequency'] = 5430e6
        radar_def[rname]['WaveLength'] = (
            c_speed/radar_def[rname]['Frequency']*1e2)

        rname = 'L'
        radar_def[rname] = radar_default.copy()
        radar_def[rname]['dwhname'] = "MLE"
        radar_def[rname]['RadarName'] = "Lema"
        radar_def[rname]['RadarCHY'] = 707957
        radar_def[rname]['RadarCHX'] = 99762
        radar_def[rname]['RadarLat'] = 46.040761
        radar_def[rname]['RadarLon'] = 8.833217
        radar_def[rname]['RadarHeight'] = 1626.0
        radar_def[rname]['ScanName'] = "1279610112"
        radar_def[rname]['Frequency'] = 5455e6
        radar_def[rname]['WaveLength'] = (
            c_speed/radar_def[rname]['Frequency']*1e2)

        rname = 'P'
        radar_def[rname] = radar_default.copy()
        radar_def[rname]['dwhname'] = "PPM"
        radar_def[rname]['RadarName'] = "PlaineMorte"
        radar_def[rname]['RadarCHY'] = 603687
        radar_def[rname]['RadarCHX'] = 135476
        radar_def[rname]['RadarLat'] = 46.370646
        radar_def[rname]['RadarLon'] = 7.486552
        radar_def[rname]['RadarHeight'] = 2937.0
        radar_def[rname]['ScanName'] = "0"
        radar_def[rname]['Frequency'] = 5468e6
        radar_def[rname]['WaveLength'] = (
            c_speed/radar_def[rname]['Frequency']*1e2)

        rname = 'W'
        radar_def[rname] = radar_default.copy()
        radar_def[rname]['dwhname'] = "WEI"
        radar_def[rname]['RadarName'] = "WeissFluhGipfel"
        radar_def[rname]['RadarCHY'] = 779700
        radar_def[rname]['RadarCHX'] = 189790
        radar_def[rname]['RadarLat'] = 47.284333
        radar_def[rname]['RadarLon'] = 9.794458
        radar_def[rname]['RadarHeight'] = 2850.0
        radar_def[rname]['ScanName'] = "0"
        radar_def[rname]['Frequency'] = 5433e6
        radar_def[rname]['WaveLength'] = (
            c_speed/radar_def[rname]['Frequency']*1e2)

    return radar_def


def _selex2deg(angle):
    """
    Convert angles from SELEX format to degree

    Parameters
    ----------
    angle : float
        angle in DN


    Returns
    -------
    conv : float
        angle in degrees

    """
    conv = angle * 360. / 65535.
    return conv



""" Unit Tests for Py-ART's io/mdv.py module. """

import bz2
from io import BytesIO

import pytest

import pyart


def test_autoread_mdv():
    radar = pyart.io.read(pyart.testing.MDV_PPI_FILE)
    source = radar.metadata['source']
    assert source == 'MDV radar volume file created by Dsr2Vol.'

    radar = pyart.io.read(pyart.testing.MDV_RHI_FILE)
    source = radar.metadata['source']
    assert source == 'MDV radar volume file created by Dsr2Vol.'


def test_autoread_sigmet():
    radar = pyart.io.read(pyart.testing.SIGMET_PPI_FILE, use_rsl=False)
    assert radar.metadata['original_container'] == 'sigmet'

    radar = pyart.io.read(pyart.testing.SIGMET_RHI_FILE, use_rsl=False)
    assert radar.metadata['original_container'] == 'sigmet'


@pytest.mark.skipif(not pyart.io.rsl._RSL_AVAILABLE,
                    reason="TRMM RSL is not installed.")
def test_autoread_sigmet_rsl():
    radar = pyart.io.read(pyart.testing.SIGMET_PPI_FILE, use_rsl=True)
    assert radar.metadata['original_container'] == 'rsl'

    radar = pyart.io.read(pyart.testing.SIGMET_RHI_FILE, use_rsl=True)
    assert radar.metadata['original_container'] == 'rsl'


def test_autoread_cfradial():
    radar = pyart.io.read(pyart.testing.CFRADIAL_PPI_FILE)
    assert radar.metadata['comment'] == 'none'

    radar = pyart.io.read(pyart.testing.CFRADIAL_RHI_FILE)
    assert radar.metadata['comment'] == 'none'


def test_autoread_nexrad_archive():
    radar = pyart.io.read(pyart.testing.NEXRAD_ARCHIVE_MSG31_COMPRESSED_FILE,
                          delay_field_loading=True)
    assert radar.metadata['original_container'] == 'NEXRAD Level II'

    radar = pyart.io.read(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE,
                          delay_field_loading=True)
    assert radar.metadata['original_container'] == 'NEXRAD Level II'


def test_autoread_nexrad_cdm():
    with pyart.testing.InTemporaryDirectory():
        tmpfile = 'tmp_nexrad.nc'
        with open(tmpfile, 'wb') as f:
            f.write(bz2.BZ2File(pyart.testing.NEXRAD_CDM_FILE).read())
        radar = pyart.io.read(tmpfile)
        assert radar.metadata['original_container'] == 'NEXRAD Level II'


def test_autoread_nexrad_level3():
    radar = pyart.io.read(pyart.testing.NEXRAD_LEVEL3_MSG19)
    assert radar.metadata['original_container'] == 'NEXRAD Level 3'


def test_autoread_raises():
    f = BytesIO(b'0000000000000000000')
    pytest.raises(TypeError, pyart.io.read, f)


headers = [
    (b'\x00\x00\x03\xf8\x00\x007>\x00\x00\x00\x01', 'MDV'),
    (b'\x89HDF\r\n\x1a\n\x02\x08\x08\x00', 'NETCDF4'),
    (b'CDF', 'NETCDF3'),
    (b'AR2V0006.501', 'WSR88D'),
    (b'SDUS54 KBMX ', 'NEXRADL3'),
    (b'\x1b\x00\x08\x00\x00\x08\xb7\x07\x00\x00\x00\x00', 'SIGMET'),
    (b'BZh91AY&SY\xbd\x12', 'BZ2'),
    (b'UF', 'UF'),                   # not from a real file
    (b'SSWB', 'DORADE'),             # not from a real file
    (b'RSL', 'RSL'),                 # not from a real file
    (b'\x0e\x03\x13\x01', 'HDF4'),   # not from a real file
    (b'000000000000', 'UNKNOWN'),
    ]
@pytest.mark.parametrize("i", headers)
def test_determine_filetype(i):
    string, filetype = i
    check_filetype.description = 'determine filetype: %s' % (filetype)
    check_filetype(string, filetype)


def check_filetype(string, filetype):
    f = BytesIO(string)
    assert pyart.io.auto_read.determine_filetype(f) == filetype
